// Twist.cpp
// Screensaver module for Hiroshi LockHeimer's saver program.
// Copyright ©1997 Michael D. Crawford.  All Rights Reserved.
// 8 Mar 97 crawford@scruznet.com http://www.scruznet.com/~crawford/
// You may use this source code as you please, provided you read this web page:
// http://www.wordservices.org/be.html

#include "SaverModule.h"
#include "BrowserLauncher.h"

void  _get_screen_bitmap_(BBitmap *,BRect,bool);

class ScreenInfo{
	public:
		ScreenInfo( BView *view, screen_id screenNum ) /*throw( int )*/;
		virtual ~ScreenInfo();

		static long SaverThreadStart( void *data );
		long RealStart();
		
		void StopRunning(){ mIsRunning = false; };
		BView *GetView(){ return mView; };
	protected:
		bool		mIsRunning;
		screen_id	mScreenNum;
		short		mCurLine;
		BBitmap		*mTempWorld;
		BBitmap		*mOldScreen;
		color_space	mColorSpace;
		BRect		mScreenRect;
		BView		*mView;
		bool		mHasDrawn;
};

class TwistInfo{
	public:		
		TwistInfo( BView *view ) /*throw( int )*/;
		virtual ~TwistInfo();
		virtual long StartThreads();
		virtual long StopThreads();
		ScreenInfo *GetScreen(){ return mScreen; };
	protected:
		ScreenInfo	*mScreen;
		thread_id	mSaverThread;
};

class ConfigHandler: public BHandler
{
	public:
		ConfigHandler();
		virtual ~ConfigHandler();
		virtual void MessageReceived( BMessage *msg );
};

class MyView: public BView
{
	public:
		MyView( bool *hasDrawn, BBitmap *map, BRect frame, const char *name, long foo, long bar );
		virtual ~MyView();
		virtual void Draw( BRect updateRect );
	protected:
		BBitmap *mMap;
		bool	*mHasDrawn;
};
	
TwistInfo *gTwistInfo;
ConfigHandler *gHandler;

// Uncomment the following line to prevent 'Random' from calling
// your module

//char inhibit_random = 1;


int main( int argc, char **argv )
{
	return 0;
}

// ------------------------------------------------------------
// 		 module_initialize  [optional]
// ------------------------------------------------------------
// This function gets called once, immediately after saver has
// loaded your module
//
// It is not known at this point whether you are being loaded
// to save the screen or to configure yourself (or both)
//
// inSettings is a pointer to data inSettingsSize long that saver 
// has stored for you, it is your responsibility to free it before 
// your module is unloaded
//
// saver may pass you NULL and 0 if it cannot find any data that
// belongs to you

void
module_initialize(
	void	*inSettings,
	long	inSettingsSize)
{
#pragma unused( inSettings )
#pragma unused( inSettingsSize )
}


// ------------------------------------------------------------
// 		 module_cleanup  [optional]
// ------------------------------------------------------------
// This function gets called once, immediately before saver will
// unload your module
//
// This is the last chance to clean up after yourself
//
// Set outSettings to a pointer that is outSettingsSize long, 
// and saver will store it for you until the next time your module 
// is loaded
//
// Pass back NULL and 0 if you do not need to store anything
//
// Do not attempt to free what you pass back to saver

void
module_cleanup(
	void	**outSettings,
	long	*outSettingsSize)
{
	*outSettings = NULL;
	outSettingsSize = 0;

	return;
}


// ------------------------------------------------------------
// 		module_start_saving  [required]
// ------------------------------------------------------------
// This function gets called when it is time to "save" the screen
//
// You are given a BView in which you are to do your work
//
// The window in which the BView resides is not shown at this time,
// it is your responsibility to inView->Window()->Show() it
//
// Spawn your thread(s) here and return from this function ASAP
// Do all your drawing in inView, or in its children (if you 
// decide to add any)
//
// inView's View and Low color are set to black, High color is white
//
// saver never locks inView's window for you, so be sure to
// inView->Window()->Lock() before you try to do anything with
// inView, otherwise you win a free trip to the debugger

void
module_start_saving(
	BView	*inView)
{
	BRect screenRect;
	color_space colorSpace;
	BBitmap *oldScreen;
	
	{
		BScreen screen( B_MAIN_SCREEN_ID ); //mScreenNum
		
		screenRect = screen.Frame();
		colorSpace = screen.ColorSpace();
	}
		
	oldScreen = new BBitmap( screenRect, colorSpace, true );
	if ( !oldScreen ){
		return;
	}

	_get_screen_bitmap_( oldScreen, screenRect, true );

	gTwistInfo = (TwistInfo*)NULL;
	gTwistInfo = new TwistInfo( inView );
	
	inView->Window()->Lock();
		
	inView->AddChild( gTwistInfo->GetScreen()->GetView() );

	inView->Window()->Show();	
	inView->Window()->Unlock();

	gTwistInfo->StartThreads();
	
	return;	
}

// ------------------------------------------------------------
// 		 module_stop_saving  [required]
// ------------------------------------------------------------
// This function gets called right before the BWindow that
// you have been drawing into gets closed
//
// Exit from all your thread(s), free any memory that you no 
// longer need, unlock anything that you locked, be good and 
// clean up after yourself

void
module_stop_saving()
{

	gTwistInfo->StopThreads();
	
	delete gTwistInfo;
}


// ------------------------------------------------------------
// 		 module_start_config  [required]
// ------------------------------------------------------------
// This function gets called when the user has selected your 
// module in the settings panel
//
// You are free to add any children to inView or modify it
// in any way (but don't even think about deleting it)
//
// inView is 241 units wide and 244 units high

void
module_start_config(
	BView *inView)
{
	if (!inView->Window()->Lock())
		return;

	gHandler = new ConfigHandler;
	
	inView->Looper()->AddHandler( gHandler );
	
	BStringView *view = NULL;
	
	view = new BStringView(BRect(20.0, 20.0, 100.0, 36.0),
								 B_EMPTY_STRING,
								 "Twist 1.0.1");
	view->SetFont( be_bold_font );
	inView->AddChild(view);
	view->SetViewColor(inView->ViewColor());
	
	view = new BStringView(BRect(30.0, 40.0, 200.0, 52.0),
								 B_EMPTY_STRING,
								 "©1997-98 Michael D. Crawford");
	inView->AddChild(view);
	view->SetViewColor(inView->ViewColor());

	view = new BStringView(BRect(30.0, 60.0, 200.0, 72.0),
								 B_EMPTY_STRING,
								 "Crawford Software Consulting");
	inView->AddChild(view);
	view->SetViewColor(inView->ViewColor());
	
	view = new BStringView(BRect(30.0, 80.0, 200.0, 92.0),
								 B_EMPTY_STRING,
								 "crawford@scruznet.com");
	inView->AddChild(view);
	view->SetViewColor(inView->ViewColor());
	
	BMessage *msg = new BMessage( BrowserLauncher::OpenURL );
	msg->AddString( "URL", "http://www.scruznet.com/~crawford/Computers/be.html" );
	
	BButton *but = new BButton(BRect(30.0, 100.0, 200.0, 120.0),
								 "SourceButton",
								 "Get Source Code",
								 msg );
	but->SetTarget( gHandler );
	
	inView->AddChild(but);

	msg = new BMessage( BrowserLauncher::OpenURL );
	msg->AddString( "URL", "http://www.scruznet.com/~crawford/crawford.html" );
	
	but = new BButton(BRect(5.0, 130.0, 225.0, 150.0),
								 "CSCButton",
								 "Crawford Software Consulting Web Page",
								 msg );
	but->SetTarget( gHandler );
	
	inView->AddChild(but);
	
	inView->Window()->Unlock();

}


// ------------------------------------------------------------
// 		 module_stop_config  [required]
// ------------------------------------------------------------
// This function gets called right before the BView that was passed
// to you in module_start_config() gets deleted
//
// Free any memory that you may have allocated, you know the game...
//
// Since your view will be deleted, it is not necessary to delete 
// the children that you added to it

void
module_stop_config()
{
	BLooper *looper = gHandler->Looper();
	looper->Lock();
		
	looper->RemoveHandler( gHandler );
	
	looper->Unlock();
	delete gHandler;
	
	return;
}

ConfigHandler::ConfigHandler()
{
	return;
}

ConfigHandler::~ConfigHandler()
{
	return;
}

void ConfigHandler::MessageReceived( BMessage *msg )
{
	switch ( msg->what ){
		case BrowserLauncher::OpenURL:
			BrowserLauncher::GoToWebPage( msg );
			break;
	}
	
	return;
}


ScreenInfo::ScreenInfo( BView *view, screen_id screenNum )
						: mScreenNum( screenNum ),
							mIsRunning( true ),
							mView( NULL ),
							mHasDrawn( false )
{

	mCurLine = 0;
	mTempWorld = (BBitmap*)NULL;
	mOldScreen = (BBitmap*)NULL;
	
	{
		BScreen screen( B_MAIN_SCREEN_ID ); //mScreenNum
		
		mScreenRect = screen.Frame();
		mColorSpace = screen.ColorSpace();
	}
		
	mOldScreen = new BBitmap( mScreenRect, mColorSpace, true );
	if ( !mOldScreen ){
		return;
	}

	_get_screen_bitmap_( mOldScreen, mScreenRect, false );

	view->Window()->Lock();
	mView = new MyView( &mHasDrawn, mOldScreen, view->Bounds(), "TwistView", 0, B_WILL_DRAW );
	view->Window()->Unlock();
	
		//BRect tempFrame( 100, 100, 200, 200 );

	BRect	tempFrame( 0, mScreenRect.top, 0, mScreenRect.bottom );
	mTempWorld = new BBitmap( tempFrame, mColorSpace, true );
	
	return;
}

ScreenInfo::~ScreenInfo()
{
	BWindow *win = mView->Window();
	win->Lock();
	mView->Parent()->RemoveChild( mView );
	win->Unlock();
	
	delete mView;

	delete mTempWorld;
	delete mOldScreen;

	
	return;
}

long ScreenInfo::SaverThreadStart( void *data )
{
	long	result;
	

	result = ( (ScreenInfo*)data )->RealStart();

	return result;
}

long ScreenInfo::RealStart( void )
{
	while ( mHasDrawn == false ){
		snooze( 1000 );
		// Sit 'n spin
	}

	while ( mIsRunning ){
		
		// Shift the pixels over by one
		BRect from( mScreenRect );
		BRect to( mScreenRect );
		
		from.right -= 1.0;
		to.left += 1.0;
		
		from.bottom = mCurLine;
		to.bottom = mCurLine;
		
		// Get the righthand column
		BRect screenFrame( mScreenRect.right, mScreenRect.top, mScreenRect.right, mScreenRect.bottom );
		BRect fromFrame( 0, mScreenRect.top, 0, mCurLine );
		BRect toFrame( 0, mScreenRect.top, 0, mCurLine );

		//BRect fromFrame( 100, 100, 200, 200 );
		//BRect toFrame( 200, 200, 300, 300 );
		
		
		mCurLine++;
		
		_get_screen_bitmap_( mTempWorld, screenFrame, true );

		if ( mCurLine > mScreenRect.bottom )
			mCurLine = 0;
		
		mView->Window()->Lock();	

		mView->CopyBits( from, to );

		// Put the lefthand column
		mView->DrawBitmap( mTempWorld, fromFrame, toFrame );
		//mView->FillRect( fromFrame );
		//mView->FillRect( toFrame );
		
		mView->Window()->Unlock();
	}

	return B_NO_ERROR;
}

TwistInfo::TwistInfo( BView *view )
{
	int	i;

	// STUB doesn't really handle multiple screens yet
		
	mScreen = new ScreenInfo( view, B_MAIN_SCREEN_ID );
	if ( !mScreen ){
		debugger("allocation failed" );
		//throw int( 0 );
		return;
	}
		
		mSaverThread = spawn_thread( ScreenInfo::SaverThreadStart,
											"Mister Twister",
											B_NORMAL_PRIORITY,
											(void*)(mScreen) );
		
		if ( mSaverThread < B_NO_ERROR ){

			//throw int( 0 );
			return;
		}
		


	return;
}

TwistInfo::~TwistInfo()
{
	int	i;

	delete mScreen;

	return;
}

long TwistInfo::StartThreads()
{
	int	i;
	
	resume_thread( mSaverThread );
	
	return B_NO_ERROR;
}

long TwistInfo::StopThreads()
{
	int	i;
	long exitValue;

	mScreen->StopRunning();

	wait_for_thread( mSaverThread, &exitValue );

	
	return B_NO_ERROR;
}

MyView::MyView( bool *hasDrawn, BBitmap *map, BRect frame, const char *name, long foo, long bar )
	: BView( frame, name, foo, bar | B_WILL_DRAW ),
	mMap( map ),
	mHasDrawn( hasDrawn )
{
	*mHasDrawn = false;
	
	// So we don't see a flash as the view is erased to white first
	// hmm doesn't seem to help
	SetViewColor( B_TRANSPARENT_32_BIT );

	return;
}

MyView::~MyView()
{
	return;
}

void MyView::Draw( BRect updateRect )
{
	DrawBitmap( mMap, Bounds(), Bounds() );
	
	*mHasDrawn = true;
	
	return;
}

